/*******************************************************************************
 * Copyright (c) 2016, 2018 Contributors to the Eclipse Foundation
 *
 * See the NOTICE file(s) distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0
 *
 * SPDX-License-Identifier: EPL-2.0
 *******************************************************************************/

package org.eclipse.hono.util;

import java.util.Optional;
import java.util.function.Function;

import org.apache.qpid.proton.amqp.messaging.ApplicationProperties;

/**
 * A container for the result returned by Hono's Tenant API.
 *
 * @param <T> The concrete type of the payload that is conveyed in the result.
 *
 */
public final class TenantResult<T> extends RequestResponseResult<T> {

    private TenantResult(
            final int status,
            final T payload,
            final CacheDirective cacheDirective,
            final ApplicationProperties applicationProperties) {
        super(status, payload, cacheDirective, applicationProperties);
    }

    /**
     * Creates a new result for a status code.
     *
     * @param status The status code indicating the outcome of the request.
     * @param <T> The type of the payload conveyed in the result.
     * @return The result.
     */
    public static <T> TenantResult<T> from(final int status) {
        return new TenantResult<>(status, null, null, null);
    }

    /**
     * Creates a new result for a status code and payload.
     *
     * @param status The status code indicating the outcome of the request.
     * @param payload The payload to convey to the sender of the request.
     * @param <T> The type of the payload conveyed in the result.
     * @return The result.
     */
    public static <T> TenantResult<T> from(final int status, final T payload) {
        return new TenantResult<>(status, payload, null, null);
    }

    /**
     * Creates a new result for a status code and payload.
     *
     * @param status The status code indicating the outcome of the request.
     * @param payload The payload to convey to the sender of the request.
     * @param <T> The type of the payload conveyed in the result.
     * @param cacheDirective Restrictions regarding the caching of the payload.
     * @return The result.
     */
    public static <T> TenantResult<T> from(final int status, final T payload, final CacheDirective cacheDirective) {
        return new TenantResult<>(status, payload, cacheDirective, null);
    }

    /**
     * Creates a new result for a status code and payload.
     *
     * @param status The status code indicating the outcome of the request.
     * @param payload The payload to convey to the sender of the request.
     * @param <T> The type of the payload conveyed in the result.
     * @param cacheDirective Restrictions regarding the caching of the payload.
     * @param applicationProperties Arbitrary properties conveyed in the response message's
     *                              <em>application-properties</em>.
     * @return The result.
     */
    public static <T> TenantResult<T> from(
            final int status,
            final T payload,
            final CacheDirective cacheDirective,
            final ApplicationProperties applicationProperties) {
        return new TenantResult<>(status, payload, cacheDirective, applicationProperties);
    }

    /**
     * Map the payload type, if set.
     *
     * @param mapper The mapper to apply, if a payload is set.
     * @param <U> The target type.
     * @return The mapped result.
     */
    public <U> TenantResult<U> map(final Function<? super T, ? extends U> mapper) {
        return new TenantResult<U>(
                getStatus(),
                Optional.ofNullable(getPayload()).map(mapper).orElse(null),
                getCacheDirective(),
                new ApplicationProperties(getApplicationProperties())
        );
    }
}
